/**************************************************************************
 *
 *  THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY
 *  KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A PARTICULAR
 *  PURPOSE.
 *
 *  Copyright (c) 1993  Microsoft Corporation.  All Rights Reserved.
 * 
 **************************************************************************/

// the usual includes

#include <windows.h>
#include <mmsystem.h>
#include <shellapi.h>
#include <commdlg.h>
#include <stdlib.h>
#include <memory.h>

// this uses GDI to copy DIB bits, instead of the routines
// in fast32.asm or slow.c
// #define USESTRETCHDIBITS

// other #define's
#define SZ_SPRITE   "Sprite"	// clipboard format name

// Raster ops used in drawing
#define DSa     0x008800C6L
#define DSx     0x00660046L
#define DSo     0x00EE0086L
#define DSna    0x00220326L

// RGB color values
#define rgbWhite    RGB(255,255,255)
#define rgbBlack    RGB(0,0,0)

// Flags used for the Redraw function
#define UPDATE_SCREEN   TRUE
#define NO_UPDATE       FALSE

// update system
//#define WAIT_LOOP
//#ifndef WAIT_LOOP
//#define TIMER_TICKS
//#endif

// Various types
typedef char huge *HPSTR;
typedef BYTE huge *HPBYTE;
typedef LPBITMAPINFO PDIB;

// the sprite structure, which goes into a sprite list
typedef struct _SPRITE
{
	struct _SPRITE FAR *pNext;  // pointer to the next item
	struct _SPRITE FAR *pPrev;  // pointer to the prev item
	PDIB pDIB;                  // The DIB image of the sprite
	int x;                      // X Coordinate of top-left corner
	int y;                      // Y Coordinate of top-left corner
	int z;                      // Z order for sprite
	int width;                  // width of bounding rectangle
	int height;                 // height of bounding rectangle
	BYTE bTopLeft;              // top left pixel value
	COLORREF rgbTopLeft;        // top left pixel color
	BOOL bVisible;				// TRUE if sprite is visible
} SPRITE, FAR *PSPRITE;

// drawrects are used for making a linked list of what on the
// screen needs to be refreshed.
typedef struct _DRAWRECT
{
	struct _DRAWRECT *pPrev;    // pointer to the prev one
	struct _DRAWRECT *pNext;    // pointer to the next one
	RECT rc;                    // the data item
} DRAWRECT, FAR*PDRAWRECT;

// this is kinda redundant, oh well...
typedef struct _DRAWRECTLIST
{
	PDRAWRECT pHead;                // pointer to the first item
} DRAWRECTLIST, FAR *PDRAWRECTLIST;

// useful bitmap macros
// macros to access the fields in a BITMAPINFO struct
// field_value = macro(pBitmapInfo)
#define BI_WIDTH(pBI)       (int)((pBI)->bmiHeader.biWidth)
#define BI_HEIGHT(pBI)      (int)((pBI)->bmiHeader.biHeight)
#define BI_PLANES(pBI)      ((pBI)->bmiHeader.biPlanes)
#define BI_BITCOUNT(pBI)    ((pBI)->bmiHeader.biBitCount)
#define BI_CLRUSED(pBI)     ((pBI)->bmiHeader.biClrUsed)

// macros to access BITMAPINFO fields in a DIB
// field_value = macro(pDIB)
#define DIB_WIDTH(pDIB)     (BI_WIDTH((LPBITMAPINFO)(pDIB)))
#define DIB_HEIGHT(pDIB)    (BI_HEIGHT((LPBITMAPINFO)(pDIB)))
#define DIB_PLANES(pDIB)    (BI_PLANES((LPBITMAPINFO)(pDIB)))
#define DIB_BITCOUNT(pDIB)  (BI_BITCOUNT((LPBITMAPINFO)(pDIB)))
#define DIB_CLRUSED(pDIB)   (BI_CLRUSED((LPBITMAPINFO)(pDIB)))
#define DIB_COLORS(pDIB)    (NumDIBColorEntries((LPBITMAPINFO)(pDIB)))
#define DIB_PCLRTAB(pDIB)   ((LPRGBQUAD)(((LPSTR)((LPBITMAPINFO)(pDIB))) \
		+ sizeof(BITMAPINFOHEADER)))
#define DIB_BISIZE(pDIB)    (sizeof(BITMAPINFOHEADER) \
	+ DIB_COLORS(pDIB) * sizeof(RGBQUAD))
#define DIB_PBITS(pDIB)     (((LPSTR)((LPBITMAPINFO)(pDIB))) \
	+ DIB_BISIZE(pDIB))
#define DIB_PBI(pDIB)       ((LPBITMAPINFO)(pDIB))
#define DIB_STORAGEWIDTH(pDIB) ((DIB_WIDTH(pDIB) + 3) & ~3)

// here's my "SpriteWindow" data struct.
// this is where all the old global variables are now;
// allows "multiple windows" of sprites.
//
// In the code it's always referenced as "pSprWnd" and sprites
// are "pSprite"s.
//
// (By the way, not all of these are used anymore...
// I just never bothered to figure out which ones)
typedef struct _SPRITESPLATZ
{
	char *szAppName;
	HINSTANCE hAppInstance;
	HWND hwndMain;
	TEXTMETRIC tmSysFont;
	HANDLE hAccTable;
	UINT uiCfSprite;
	PSPRITE pSpriteList;
	PDIB pdibBkGnd;
	PDIB pdibOffScreen;
	HDC hdcOffScreen;
	LPBITMAPINFO pPalClrTable;
	UINT uiTimer;
	HPALETTE hpalCurrent;
	BOOL bShowUpdateRects;
	char szIniFile[_MAX_PATH];
	BOOL bAutoUpdate;
	DRAWRECTLIST DrawList;
} SPRITEWND, FAR *PSPRITEWND;

// these are still global (to the DLL, that is).
int __iDebugLevel;
HANDLE __hMem;

// useful macros for handling memory...
#define ALLOCATE(s) ((__hMem = GlobalAlloc(GHND, (DWORD)(s))) ? GlobalLock(__hMem) : NULL)
#define FREE(p) (GlobalUnlock((HANDLE)HIWORD(p)),GlobalFree((HANDLE)HIWORD(p)))
#define SIZE(p) (GlobalSize((HANDLE)HIWORD(p)))

// and, last but not least, forward declarations for all the functions.
// (some of these functions don't exist any more...)
// main.c
extern UINT cdecl Message(UINT uiBtns, LPSTR lpFormat, ...) ;
// init.c
extern BOOL InitFirstInstance(HANDLE hInstance);
extern BOOL InitCurrentInstance(HANDLE hInstance, LPSTR lpszCmdLine, int cmdShow);
extern void Terminate(void);
extern void LoadScene(LPSTR pFileName);
// help.c
extern void About(HWND hWnd);
extern void Help(HWND hWnd, WORD wParam);
// file.c
#define PFFN_OPENFILE        0x0001
#define PFFN_SAVEFILE        0x0002
#define PFFN_SHOWDEFAULT     0x0004
#define PFFN_OVERWRITEPROMPT 0x0008
#define PFFN_UPPERCASE       0x0010
extern BOOL PromptForFileName(HWND hwndOwner,
	HANDLE hInst,
	LPSTR achFileName,
	WORD cchFileName,
	LPSTR szCaption,
	LPSTR szFilter,
	LPSTR szDefExt,
	DWORD dwFlags);
// print.c
extern void Print(HDC hDC);
extern LRESULT CALLBACK DebugWndProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam);

// if debugging is on, send stuff out via DbgOut
#ifdef DEBUG
DWORD   __dwEval;
extern void SetDebugLevel(int i);
extern void __AssertMsg(LPSTR exp ,LPSTR file, int line);
extern void cdecl DbgOut(LPSTR lpFormat, ...);
#define ASSERT(exp) \
((exp) ? (void)0 : __AssertMsg((LPSTR)(#exp), (LPSTR)__FILE__, __LINE__))
#define EVAL(exp) \
(__dwEval=(DWORD)(exp), (__dwEval) ? (void)0 : __AssertMsg((LPSTR)(#exp), (LPSTR)__FILE__, __LINE__), __dwEval)
#define dprintf                          DbgOut
#define dprintf1 if (__iDebugLevel >= 1) DbgOut
#define dprintf2 if (__iDebugLevel >= 2) DbgOut
#define dprintf3 if (__iDebugLevel >= 3) DbgOut
#define dprintf4 if (__iDebugLevel >= 4) DbgOut
#else
#define ASSERT(exp) 0
#define EVAL(exp) (exp)
#define dprintf  if (0) ((int (*)(char *, ...)) 0)
#define dprintf1 if (0) ((int (*)(char *, ...)) 0)
#define dprintf2 if (0) ((int (*)(char *, ...)) 0)
#define dprintf3 if (0) ((int (*)(char *, ...)) 0)
#define dprintf4 if (0) ((int (*)(char *, ...)) 0)
#endif

// more forward declares (most of these do still exist).
// bkgnd.c
// palette.c
extern HPALETTE CreateDIBPalette(LPBITMAPINFO lpBmpInfo);
extern LRESULT PaletteMessage(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam);
extern HPALETTE CopyPalette(HPALETTE hSrcPal);
extern void SetSysPalColors(HPALETTE hPal);
// draw.c
extern void RenderDIBOffScreen(PSPRITEWND pSprWnd, PDIB pDIB, int iX, int iY, LPRECT prcClip);
extern void Paint(PSPRITEWND pSprWnd, HDC hDC, LPRECT prcClip);
extern void FillBkGnd(PSPRITEWND pSprWnd);
extern void Redraw(PSPRITEWND pSprWnd, LPRECT prcClip, BOOL bUpdateScreen);
extern PDRAWRECTLIST CreateDrawRectList(void);
extern void DeleteDrawRectListItem(PDRAWRECTLIST pList, PDRAWRECT pItem);
extern void EmptyDrawRectList(PDRAWRECTLIST pList);
extern void AddDrawRectItem(PDRAWRECTLIST pList, LPRECT pRect);
extern void MergeDrawRectList(PDRAWRECTLIST pList);
// sprite.c
extern PSPRITE PASCAL _export SpriteHitTest(PSPRITEWND pSprWnd, int x, int y);
extern void PASCAL _export SetSpritePosition(PSPRITEWND pSprWnd, PSPRITE pSprite, int x, int y, BOOL bUpdate);
extern void GetSpriteRect(PSPRITE pSprite, LPRECT prcSprite);
extern void PASCAL _export DeleteSprite(PSPRITEWND pSprWnd, PSPRITE pSprite);
extern void DeleteSpriteList(PSPRITEWND pSprWnd);
extern void PASCAL _export SetSpriteZOrder(PSPRITEWND pSprWnd, PSPRITE pSprite, WORD z, BOOL bUpdate);
// dib.c
extern PDIB PASCAL _export LoadDIB(LPSTR pszPath);
extern void PASCAL _export DeleteDIB(PDIB pDIB);
extern BYTE GetDIBPixelValue(PDIB pDIB, int x, int y);
extern COLORREF GetDIBPixelColor(PDIB pDIB, int x, int y);
extern BOOL IsWinDIB(LPBITMAPINFOHEADER pBI);
extern void ShowInfo(LPBITMAPINFO lpBmpInfo);
extern WORD NumDIBColorEntries(LPBITMAPINFO lpBmpInfo);
extern PDIB CreateCompatibleDIB(PDIB pOld);
extern HPSTR GetDIBPixelAddress(PDIB pDIB, int x, int y);
extern void MapDIBColorTable(PDIB pdibObj, PDIB pdibRef);
// fast.asm or slow.c
extern void FAR PASCAL TransCopyDIBBits(HPBYTE pD,
	HPBYTE pS,
	DWORD dwWidth,
	DWORD dwHeight,
	DWORD dwScanD,
	DWORD dwScanS,
	BYTE bTranClr);
extern void FAR PASCAL CopyDIBBits(HPBYTE pD,
	HPBYTE pS,
	DWORD dwWidth,
	DWORD dwHeight,
	DWORD dwScanD,
	DWORD dwScanS);
